@@ -16,12 +16,15 @@
 //!
 //! ### Note: Windows Compatibility
 //!
-//! [`OsStr`] and related structs will be printed lossily on Windows. Paths are
+//! [`OsStr`] and related structs may be printed lossily on Windows. Paths are
 //! not represented using bytes on that platform, so it may be confusing to
-//! print them in that manner. Plus, the encoding most often used to account
+//! display them in that manner. Plus, the encoding most often used to account
 //! for the difference is [not permitted to be written to files][wtf-8
 //! audience], so it would not make sense for this crate to use it.
 //!
+//! Windows Console can display these paths, so this crate will output them
+//! losslessly when writing to that terminal.
+//!
 //! # Features
 //!
 //! These features are optional and can be enabled or disabled in a
@@ -41,11 +44,10 @@
 //!
 //! ```
 //! use std::env;
-//! # use std::io::Result as IoResult;
+//! # use std::io;
 //!
 //! use print_bytes::println_bytes;
 //!
-//! # fn main() -> IoResult<()> {
 //! print!("exe: ");
 //! println_bytes(&env::current_exe()?);
 //! println!();
@@ -54,8 +56,8 @@
 //! for arg in env::args_os().skip(1) {
 //!     println_bytes(&arg);
 //! }
-//! #     Ok(())
-//! # }
+//! #
+//! # Ok::<_, io::Error>(())
 //! ```
 //!
 //! [array]: https://doc.rust-lang.org/std/primitive.array.html
@@ -69,19 +71,19 @@
 //! [wtf-8 audience]: https://simonsapin.github.io/wtf-8/#intended-audience
 
 #![cfg_attr(
-    any(all(doc, not(doctest)), feature = "const_generics"),
+    any(feature = "const_generics", feature = "specialization"),
     allow(incomplete_features)
 )]
 #![doc(html_root_url = "https://docs.rs/print_bytes/*")]
-#![cfg_attr(
-    any(all(doc, not(doctest)), feature = "const_generics"),
-    feature(const_generics)
-)]
+#![cfg_attr(any(feature = "const_generics"), feature(const_generics))]
+// Only require a nightly compiler when building documentation for docs.rs.
+// This is a private option that should not be used.
+// https://github.com/rust-lang/docs.rs/issues/147#issuecomment-389544407
+#![cfg_attr(print_bytes_docs_rs, feature(doc_cfg))]
 #![cfg_attr(feature = "specialization", feature(specialization))]
 #![warn(unused_results)]
 
 use std::io;
-use std::io::Result as IoResult;
 use std::io::Stderr;
 use std::io::StderrLock;
 use std::io::Stdout;
@@ -90,24 +92,38 @@ use std::io::Write;
 
 mod bytes;
 pub use bytes::Bytes;
+use bytes::BytesInner;
 pub use bytes::ToBytes;
 
-#[cfg(unix)]
-#[path = "unix.rs"]
+#[cfg(not(windows))]
+#[path = "common.rs"]
 mod imp;
 #[cfg(windows)]
 #[path = "windows.rs"]
 mod imp;
 
 trait WriteBytes: Write {
-    fn is_console(&self) -> bool;
+    fn to_console(&self) -> Option<imp::Console<'_>>;
 
     #[inline]
-    fn write_bytes<'a, TValue>(&mut self, value: &'a TValue) -> IoResult<()>
+    fn write_bytes<TValue>(&mut self, value: &TValue) -> io::Result<()>
     where
-        TValue: ?Sized + ToBytes<'a>,
+        TValue: ?Sized + ToBytes,
     {
-        imp::write(self, &value.to_bytes().0)
+        let value = value.to_bytes().0;
+        match value {
+            BytesInner::Bytes(value) => {
+                let buffer;
+                let value = if self.to_console().is_some() {
+                    buffer = String::from_utf8_lossy(value);
+                    buffer.as_bytes()
+                } else {
+                    &value
+                };
+                self.write_all(value)
+            }
+            BytesInner::OsStr(value) => imp::write_os(self, value),
+        }
     }
 }
 
@@ -116,8 +132,19 @@ impl<T> WriteBytes for T
 where
     T: ?Sized + Write,
 {
-    default fn is_console(&self) -> bool {
-        false
+    default fn to_console(&self) -> Option<imp::Console<'_>> {
+        None
+    }
+}
+
+#[cfg(feature = "specialization")]
+impl<'a, T> WriteBytes for &'a mut T
+where
+    T: ?Sized + WriteBytes,
+    &'a mut T: Write,
+{
+    default fn to_console(&self) -> Option<imp::Console<'_>> {
+        (**self).to_console()
     }
 }
 
@@ -125,8 +152,8 @@ macro_rules! r#impl {
     ( $($type:ty),* $(,)? ) => {
         $(
             impl WriteBytes for $type {
-                fn is_console(&self) -> bool {
-                    imp::is_console(self)
+                fn to_console(&self) -> Option<imp::Console<'_>> {
+                    imp::Console::from_handle(self)
                 }
             }
         )*
@@ -141,23 +168,22 @@ r#impl!(Stderr, StderrLock<'_>, Stdout, StdoutLock<'_>);
 ///
 /// For more information, see [the module-level documentation][module].
 ///
-/// *This function is only available with the `"specialization"` feature.*
-///
 /// # Errors
 ///
 /// Returns an error if writing fails.
 ///
 /// [module]: index.html
 /// [`write!`]: https://doc.rust-lang.org/std/macro.write.html
-#[cfg(any(doc, feature = "specialization"))]
+#[cfg_attr(print_bytes_docs_rs, doc(cfg(feature = "specialization")))]
+#[cfg(feature = "specialization")]
 #[inline]
-pub fn write_bytes<'a, TValue, TWriter>(
-    writer: &mut TWriter,
-    value: &'a TValue,
-) -> IoResult<()>
+pub fn write_bytes<TValue, TWriter>(
+    mut writer: TWriter,
+    value: &TValue,
+) -> io::Result<()>
 where
-    TValue: ?Sized + ToBytes<'a>,
-    TWriter: ?Sized + Write,
+    TValue: ?Sized + ToBytes,
+    TWriter: Write,
 {
     writer.write_bytes(value)
 }
@@ -171,9 +197,9 @@ macro_rules! r#impl {
     ) => {
         $(#[$print_fn_attr])*
         #[inline]
-        pub fn $print_fn<'a, TValue>(value: &'a TValue)
+        pub fn $print_fn<TValue>(value: &TValue)
         where
-            TValue: ?Sized + ToBytes<'a>,
+            TValue: ?Sized + ToBytes,
         {
             if let Err(error) = $writer.write_bytes(value) {
                 panic!("failed writing to {}: {}", $label, error);
@@ -182,13 +208,13 @@ macro_rules! r#impl {
 
         $(#[$println_fn_attr])*
         #[inline]
-        pub fn $println_fn<'a, TValue>(value: &'a TValue)
+        pub fn $println_fn<TValue>(value: &TValue)
         where
-            TValue: ?Sized + ToBytes<'a>,
+            TValue: ?Sized + ToBytes,
         {
             let _ = $writer.lock();
             $print_fn(value);
-            $print_fn(b"\n".as_ref());
+            $print_fn(&b"\n"[..]);
         }
     };
 }
@@ -261,14 +287,15 @@ r#impl!(
 
 #[cfg(test)]
 mod tests {
-    use std::io::Result as IoResult;
+    use std::io;
     use std::io::Write;
 
     use super::imp;
     use super::WriteBytes;
 
-    const INVALID_STRING: &[u8] = b"\xF1foo\xF1\x80bar\xF1\x80\x80baz";
+    const INVALID_STRING: &[u8] = b"\xF1foo\xF1\x80bar\xF1\x80\x80";
 
+    #[derive(Debug)]
     struct Writer {
         buffer: Vec<u8>,
         is_console: bool,
@@ -284,55 +311,63 @@ mod tests {
     }
 
     impl Write for Writer {
-        fn flush(&mut self) -> IoResult<()> {
+        fn flush(&mut self) -> io::Result<()> {
             self.buffer.flush()
         }
 
-        fn write(&mut self, bytes: &[u8]) -> IoResult<usize> {
+        fn write(&mut self, bytes: &[u8]) -> io::Result<usize> {
             self.buffer.write(bytes)
         }
     }
 
     impl WriteBytes for Writer {
-        fn is_console(&self) -> bool {
-            self.is_console
+        fn to_console(&self) -> Option<imp::Console<'_>> {
+            if self.is_console {
+                // SAFETY: Since no platform strings are being written, no test
+                // should ever write to this console.
+                Some(unsafe { imp::Console::null() })
+            } else {
+                None
+            }
         }
     }
 
     fn assert_invalid_string(writer: &Writer, lossy: bool) {
-        let bytes = writer.buffer.as_slice();
-        assert_ne!(lossy, INVALID_STRING == bytes);
+        let lossy_string = String::from_utf8_lossy(INVALID_STRING);
+        let lossy_string = lossy_string.as_bytes();
+        assert_ne!(INVALID_STRING, lossy_string);
+
+        let string = writer.buffer.as_slice();
         if lossy {
-            assert_eq!(
-                String::from_utf8_lossy(INVALID_STRING).as_bytes(),
-                bytes,
-            );
+            assert_eq!(lossy_string, string);
+        } else {
+            assert_eq!(INVALID_STRING, string);
         }
     }
 
-    fn test<TWriteFn>(mut write_fn: TWriteFn) -> IoResult<()>
+    fn test<TWriteFn>(mut write_fn: TWriteFn) -> io::Result<()>
     where
-        TWriteFn: FnMut(&mut Writer, &[u8]) -> IoResult<()>,
+        TWriteFn: FnMut(&mut Writer, &[u8]) -> io::Result<()>,
     {
-        let mut writer = Writer::new(true);
-        write_fn(&mut writer, INVALID_STRING)?;
-        assert_invalid_string(&writer, cfg!(windows));
-
-        writer = Writer::new(false);
+        let mut writer = Writer::new(false);
         write_fn(&mut writer, INVALID_STRING)?;
         assert_invalid_string(&writer, false);
 
+        writer = Writer::new(true);
+        write_fn(&mut writer, INVALID_STRING)?;
+        assert_invalid_string(&writer, true);
+
         Ok(())
     }
 
     #[test]
-    fn test_write() -> IoResult<()> {
-        test(imp::write)
+    fn test_write() -> io::Result<()> {
+        test(WriteBytes::write_bytes)
     }
 
     #[cfg(feature = "specialization")]
     #[test]
-    fn test_write_bytes() -> IoResult<()> {
+    fn test_write_bytes() -> io::Result<()> {
         test(|writer, bytes| super::write_bytes(writer, bytes))
     }
 }
